(*Author: Christian Sternagel*)

section \<open>Exercises of Functional Programming Lecture 7\<close>

theory Exercises
  imports Main
begin

-- \<open>Avoid qualified names in output (which would be used whenever names already exist in Main).\<close>
declare [[names_short]]

-- \<open>A custom data type for lists.\<close>
datatype 'a list = Nil | Cons 'a "'a list"

fun app :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
  where
    "app Nil ys = ys"
  | "app (Cons x xs) ys = Cons x (app xs ys)"

fun length :: "'a list \<Rightarrow> nat"
  where
    "length Nil = 0"
  | "length (Cons _ xs) = length xs + 1"


subsection \<open>Exercise 1: The empty list is a right identity of append\<close>

lemma app_Nil:
  "app xs Nil = xs"
proof (induct xs)
  case Nil
  then show ?case by simp
next
  case IH: (Cons x xs)
  have "app (Cons x xs) Nil = Cons x (app xs Nil)" by simp
  also have "Cons x (app xs Nil) = Cons x xs" by (simp add: IH)
  finally show ?case by simp
qed


subsection \<open>Exercise 2: Append is associative\<close>

lemma app_assoc:
  "app (app xs ys) zs = app xs (app ys zs)"
proof (induct xs)
  case Nil
  then show ?case by simp
next
  case IH: (Cons x xs)
  have "app (app (Cons x xs) ys) zs = app (Cons x (app xs ys)) zs" by simp
  also have "app (Cons x (app xs ys)) zs = Cons x (app (app xs ys) zs)" by simp
  also have "Cons x (app (app xs ys) zs) = Cons x (app xs (app ys zs))" by (simp add: IH)
  also have "Cons x (app xs (app ys zs)) = app (Cons x xs) (app ys zs)" by simp
  finally show ?case by simp
qed


subsection \<open>Exercise 3: Length of append\<close>

lemma length_app:
  "length (app xs ys) = length xs + length ys"
proof (induct xs)
  case Nil
  then show ?case by simp
next
  case IH: (Cons x xs)
  have "length (app (Cons x xs) ys) = length (Cons x (app xs ys))" by simp
  also have "\<dots> = length (app xs ys) + 1" by simp
  also have "\<dots> = length xs + length ys + 1" by (simp add: IH)
  also have "\<dots> = length (Cons x xs) + length ys" by simp
  finally show ?case by simp
qed


subsection \<open>Exercise 4: Number of nodes of perfect binary tree\<close>

datatype 'a btree = Empty | Node 'a "'a btree" "'a btree"

fun sz :: "'a btree \<Rightarrow> nat"
  where
    "sz Empty = 0"
  | "sz (Node _ l r) = sz l + sz r + 1"

fun ht :: "'a btree \<Rightarrow> nat"
  where
    "ht Empty = 0"
  | "ht (Node _ l r) = max (ht l) (ht r) + 1"

fun perfect :: "'a btree \<Rightarrow> bool"
  where
    "perfect Empty \<longleftrightarrow> True"
  | "perfect (Node _ l r) \<longleftrightarrow> ht l = ht r \<and> perfect l \<and> perfect r"

lemma size_of_perfect_btree:
  "perfect t \<longrightarrow> sz t = 2 ^ (ht t) - 1"
proof (induct t)
  case Empty
  then show ?case by simp
next
  case IH: (Node x l r)
  show ?case
  proof
    assume "perfect (Node x l r)"
    then have ht_l: "ht l = ht r"
      and sz_l: "sz l = 2 ^ (ht l) - 1"
      and sz_r: "sz r = 2 ^ (ht r) - 1" by (simp add: IH)+
    have "2 ^ (ht (Node x l r)) - 1 = 2 ^ (max (ht l) (ht r) + 1) - 1" by simp
    also have "\<dots> = 2 ^ (max (ht r) (ht r) + 1) - 1" by (simp add: ht_l)
    also have "\<dots> = 2 ^ (ht r + 1) - 1" by simp
    also have "\<dots> = 2 * 2 ^ (ht r) - 1" by simp
    also have "\<dots> = 2 ^ (ht r) + 2 ^ (ht r) - (1::nat)" by simp
    also have "\<dots> = 2 ^ (ht l) + 2 ^ (ht r) - 1" by (simp add: ht_l)
    also have "\<dots> = 2 ^ (ht l) + 1 - 1 + 2 ^ (ht r) - 1" by simp
    also have "\<dots> = (2 ^ (ht l) - 1) + 1 + (2 ^ (ht r) - 1)" by simp
    also have "\<dots> = (2 ^ (ht l) - 1) + (2 ^ (ht r) - 1) + 1" by arith
    also have "\<dots> = sz l + sz r + 1" by (simp add: sz_l sz_r)
    also have "\<dots> = sz (Node x l r)" by simp
    finally show "sz (Node x l r) = 2 ^ (ht (Node x l r)) - 1" by simp
  qed
qed

end
